/* newmesh.h

    Emma Robinson and Matthew Webster, FMRIB Image Analysis Group

    Copyright (C) 2012 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/
/*  THIS CLASS CONTAINS ALL FUNCTIONS FOR READING, WRITING AND STORING SURFACES
    AS WELL AS DATA HELD ON THE SURFACE*/
#if !defined (NEWMESH_H)
#define NEWMESH_H


#define NEW_ASCII 5
#define NEW_VTK   6
#define NEW_GIFTI 7
#define NEW_MATRIX 8
#define NEW_DPV 9

#include <stdio.h>
#include <iomanip>
#include <memory>

#include "miscmaths/miscmaths.h"
#include "fslsurface/fslsurfaceio.h"
#include "fslsurface/fslsurface.h"
#include "point.h"
#include "giftiInterface.h"


// ER Currently code does not fully implement tvalues ///////////////////////////
#define EPSILON    (1.0E-8)

namespace NEWMESH{

  int  meshFileType(const std::string& filename);
  bool meshExists(const std::string& filename);

  class NEWMESHException : public std::exception{

  public:
    const char* errmesg;
    NEWMESHException(const char* msg)
    {
      errmesg=msg;
    }
    virtual const char* what() const throw()
    {
      std::cout<<errmesg<<std::endl;
      return errmesg;
    }
    //	private:

  };

  class newmesh;

  /* MPOINT STORES THE VERTEX INFORMATION, THE INDICES OF
     THEIR NEIGHBOURS AND ADJOINING FACES */
  class Mpoint {

  public:

    Mpoint(){ _coord=Pt(0, 0, 0); _no=0;};
    Mpoint(double x, double y, double z, int counter):_no(counter){ _coord=Pt(x, y, z); }
    Mpoint(const Pt p, int counter,float val=0):_no(counter){ _coord=p;}
    ~Mpoint(){}
    Mpoint(const Mpoint &p):_coord(p._coord),_no(p._no),_trID(p._trID),_nID(p._nID){*this=p;}

    Mpoint& operator=(const Mpoint& p){
      _coord=p._coord;
      _no=p._no;
      _trID=p._trID;
      _nID=p._nID;
      return(*this);
    }

    friend class newmesh;

    const Pt&    get_coord() const         {return _coord;}
    void         set_coord(const Pt& coord){ _coord=coord;}
    const int&   get_no() const            {return _no;}

    void         push_triangle(const int& i){_trID.push_back(i);}
    void         push_neighbour(const int& i){_nID.push_back(i);}

    int          ntriangles()const{return (int)_trID.size();}
    int          nneighbours()const{return (int)_nID.size();}

    const int&   get_trID(const int i)const{return _trID[i];}
    const int&   get_nID(const int i) const {return _nID[i];}

    inline bool is_neighbour(const int& );  // ER added, used as part of push_triangle
    bool is_triangle(const int& );

  private:

    Pt                  _coord;  /// vertex coordinate
    int                 _no;    /// counting index of vertex (where it appears in the _points array of mesh)
    std::vector<int>    _trID;  //// faces adjacent to vertex
    std::vector<int>    _nID;  /// neighbours of this vertex

    void         normalize(){_coord.normalize();}  //  used in make mesh_mesh_from_icosa() - made private to prevent accidental use in place of Pt::normalize elsewhere in the code

  };

  class Triangle {
    /* TRIANGLE STORES THE FACE INFORMATION: ARRAY OF ALL ADJOINING VERTEX MPOINTS */
  private:
    std::vector<std::shared_ptr<Mpoint> > _vertice;
    int               _no;
    double            _area;

    Triangle(std::shared_ptr<Mpoint> p1, std::shared_ptr<Mpoint> p2, std::shared_ptr<Mpoint> p3,int no):_no(no)
    { _vertice.clear(); _vertice.push_back(p1); _vertice.push_back(p2); _vertice.push_back(p3);
      _area=area();
    }  /// newmesh should initialise the Mpoints these point to, want to avoid any chance that 2 sets of vertices are created - so made constructor private also, and newmesh a friend class.


  public:

    Triangle(){}

    ~Triangle(){}

    friend class newmesh;

    Triangle(const Triangle &t):_vertice(t._vertice),_no(t._no),_area(t._area) { *this=t;}

    Triangle(Pt p1,Pt p2, Pt p3,int no):_no(no)
    {
      std::shared_ptr<Mpoint>  m1=std::shared_ptr<Mpoint> ( new Mpoint (p1,0));
      std::shared_ptr<Mpoint>  m2=std::shared_ptr<Mpoint> ( new Mpoint (p2,1));
      std::shared_ptr<Mpoint>  m3=std::shared_ptr<Mpoint> ( new Mpoint (p3,2));
      _vertice.clear(); _vertice.push_back(m1); _vertice.push_back(m2); _vertice.push_back(m3);

      _area=area();

    }

    Triangle& operator=(const Triangle& t){
      _vertice=t._vertice;_no=t._no;return *this;
    }

    Triangle copy() const;

    void set(Pt p1,Pt p2, Pt p3,int no)
    {
      std::shared_ptr<Mpoint>  m1=std::shared_ptr<Mpoint> ( new Mpoint (p1,0));
      std::shared_ptr<Mpoint>  m2=std::shared_ptr<Mpoint> ( new Mpoint (p2,1));
      std::shared_ptr<Mpoint>  m3=std::shared_ptr<Mpoint> ( new Mpoint (p3,2));
      _vertice.clear(); _vertice.push_back(m1); _vertice.push_back(m2); _vertice.push_back(m3);
      _no=no;
      _area=area();

    }

    Pt centroid() const{
      Pt p ((_vertice[0]->get_coord().X +_vertice[1]->get_coord().X +_vertice[2]->get_coord().X)/3,
            (_vertice[0]->get_coord().Y +_vertice[1]->get_coord().Y +_vertice[2]->get_coord().Y)/3,
            (_vertice[0]->get_coord().Z +_vertice[1]->get_coord().Z +_vertice[2]->get_coord().Z)/3);

      return p;
    }

    Pt normal() const{
      Pt result=(_vertice[2]->get_coord()-_vertice[0]->get_coord())*(_vertice[1]->get_coord()-_vertice[0]->get_coord());
      result.normalize();
      return result;
    }


    double area() const{
      Pt result=(_vertice[2]->get_coord()-_vertice[0]->get_coord())*(_vertice[1]->get_coord()-_vertice[0]->get_coord());

      return 0.5*result.norm();
    }

    inline void swap(){   //changes triangle orientation -added from mesh as part of make_mesh_from_icosa()
      std::shared_ptr<Mpoint> p = _vertice[1];
      _vertice[1] = _vertice[2];
      _vertice[2] = p;
    }

    bool isinside(const Pt& x)const;
    double dist_to_point(const Pt& x0)const;

    const double get_area() const{return _area;}

    const Mpoint& get_vertice(const int& i) const{return *_vertice[i];}

    void set_vertice(const int& i, const Pt & p) {std::shared_ptr<Mpoint>  m=std::shared_ptr<Mpoint> ( new Mpoint (p,0));
      if(i>=3) throw NEWMESHException("index exceeds triangle  dimensions");
      if(_vertice.size()==0) _vertice.resize(3);
      _vertice[i]=m;}

    const Pt& get_vertex_coord(const int& i) const{return   _vertice[i]->get_coord();}  // ER added
    const bool intersect(const std::vector<Pt> & p)const;          // checks if a segment intersects the triangle
    const bool intersect(const std::vector<Pt> & p,int& ind)const; // checks if a segment intersects the triangle+gives the index of the closest vertex
    const int get_no()const{return _no;}
    const std::vector<double> get_angles() const;
    const int get_vertex_no(const int& i)const{return _vertice[i]->get_no();} // ER added
  };


  const bool operator ==(const NEWMESH::Mpoint &p2, const NEWMESH::Mpoint &p1);

  const bool operator ==(const NEWMESH::Mpoint &p2, const NEWMESH::Pt &p1);

  const Pt operator -(const NEWMESH::Mpoint &p1, const NEWMESH::Mpoint &p2);

  const Pt operator -(const NEWMESH::Pt &p1, const NEWMESH::Mpoint &p2);

  const bool operator ==(const NEWMESH::newmesh &M1, const NEWMESH::newmesh &M2);

  class newmesh {

  private:

    std::vector<std::shared_ptr<Mpoint> >   _points;
    std::vector<NEWMESH::Triangle> _triangles;
    std::vector<Pt> _normals;

    std::vector<std::vector<float> >       _pvalues;  // enables multivariate indexed as _pvalues[feat][vertex]
    std::vector<std::vector<float> >      _tvalues;

    ////////////// META DATA /////////////////////////  note might benefit from allowing user to define some attributes, especially when loading an ascii and outputting a surf.gii?
    std::vector<GIFTImeta> global_metaData;
    std::vector<GIFTImeta> global_Attributes; //Any other attributes that the tag has
    std::map<int,GIFTIlabel> global_GIFTIlabels;
    std::vector<GIFTIcoordinateSystem> global_defaultcoord;


    void push_triangle(const Triangle& t); // made private to try to safeguared initialisation of the vertex coordinates to points
    void push_point(std::shared_ptr<Mpoint> m){_points.push_back(m);}

  public:


    newmesh();

    newmesh(const newmesh &m);

    newmesh& operator=(const newmesh& m);

    void copy_meta(const newmesh& m);

    void clear(){
      _points.clear();_triangles.clear();_pvalues.clear();_tvalues.clear();
    }

    void clear_data(){
      _pvalues.clear();_tvalues.clear();
    }

    void make_mesh_from_icosa(int n);  /// functions for generatating a regularly spaced icosahedron of order n
    int get_ico_resolution() const ;

    std::vector<unsigned int> cluster(const double threshold,const int field=-1);
    void retessellate();
    void retessellate(std::vector<std::vector<int> > &);
    //------------ACCESS-------------------------------------//

    int   nvertices() const{return (int)_points.size();}
    int   ntriangles() const{return (int)_triangles.size();}
    int   npvalues(int dim=0) const{return (int)_pvalues[dim].size();}

    const NEWMESH::Mpoint&    get_point(const int& n)const{return *_points[n];}
    const NEWMESH::Pt&    get_coord(const int& n)const{if(n>=(int) _points.size() || (int) _points.size()==0){throw NEWMESHException("get_coord: index exceeds data dimensions");} return  _points[n]->get_coord();}
    const NEWMESH::Pt&    get_normal(const int& n)const {if((int)_normals.size()<(int) _points.size()){throw NEWMESHException("get_normal: normals have not been calculated, apply estimate_normals() first");}    return _normals[n];
    }
    const NEWMESH::Triangle&  get_triangle(const int& n)const{if(n>= (int) _triangles.size() || (int) _triangles.size()==0){throw NEWMESHException("get_triangle: index exceeds face dimensions");}else return  _triangles[n];}
    const double  get_triangle_area(const int& n)const{if(n>= (int) _triangles.size() ||(int) _triangles.size()==0){throw NEWMESHException("get_triangle: index exceeds face dimensions");} else return  _triangles[n].get_area();}
    const Pt  get_triangle_normal(const int& n)const{if(n>=(int) _triangles.size() || (int) _triangles.size()==0){throw NEWMESHException("get_triangle: index exceeds face dimensions");} else return  _triangles[n].normal();}

    const int get_triangleID_from_vertexIDs(const int & n0, const int & n1,const int & n2)const;

    const NEWMESH::Triangle& get_triangle_from_vertex(const int & n, const int & ID)const{if(n>=(int) _triangles.size() || (int) _triangles.size()==0){throw NEWMESHException("get_triangle: index exceeds face dimensions");} else return _triangles[_points[n]->get_trID(ID)];}
    const NEWMESH::Pt&  get_triangle_vertex(const int& n, const int& i)const{if(n>=(int) _triangles.size() || (int) _triangles.size()==0){throw NEWMESHException("get_triangle: index exceeds face dimensions");} else return  _triangles[n].get_vertex_coord(i);}
    const int get_neighbour(const int & n, const int & ID)const{if(n>=(int) _points.size() || (int) _points.size()==0){throw NEWMESHException("get_coord: index exceeds data dimensions");} return _points[n]->get_nID(ID);}
    int  get_triangle_vertexID(const int& n, const int& i)const{if(n>=(int) _triangles.size() || (int) _triangles.size()==0){throw NEWMESHException("get_triangle: index exceeds face dimensions");} else return  _triangles[n].get_vertex_no(i);}

    bool is_triangle(const int &n, const int & ID)const{if(n>=(int) _points.size() || (int) _points.size()==0){throw NEWMESHException("get_coord: index exceeds data dimensions");} return _points[n]->is_triangle(ID);}

    float get_pvalue(const int& i, const int dim =0)const{if((int)_pvalues.size()<=dim || (int)_pvalues[dim].size()<i){throw NEWMESHException("get_pvalue: index exceeds data dimensions");} return _pvalues[dim][i];}
    float get_tvalue(const int& i, const int dim =0)const{if((int)_tvalues.size()<=dim || (int) _tvalues[dim].size()<i){throw NEWMESHException("get_tvalue: index exceeds data dimensions");} return _tvalues[dim][i];}
    int get_dimension()const{ return _pvalues.size();}

    int get_total_neighbours(const int& i)const{if(i>=(int) _points.size() || (int) _points.size()==0){throw NEWMESHException("get_coord: index exceeds data dimensions");} return _points[i]->nneighbours();}
    int get_total_triangles(const int& i)const{if(i>= (int) _points.size() || (int) _points.size()==0){throw NEWMESHException("get_coord: index exceeds data dimensions");} return _points[i]->ntriangles();}

    NEWMAT::Matrix get_pvalues()const;
    std::vector<std::shared_ptr<Mpoint> > get_points()const{return _points; };


    Pt local_normal(const int& pt)const;

    const std::vector<std::vector<double> > get_face_angles() const;

    std::vector< float > getPointsAsVectors()const;  /// accessing as vectors is used during write
    std::vector< float > getValuesAsVectors(const int &d=0)const;

    std::vector< std::vector<unsigned int> > getTrianglesAsVectors()const;
    std::vector<int> getTrianglesAsVector()const;

    ///////////////// for identfying which face (and therefore vertices) surround a given point //////////////////////////////////////////////
    NEWMAT::ReturnMatrix closest_triangle(int , const NEWMESH::Pt&) const ;
    int closest_triangle_ID(int , const NEWMESH::Pt&) const ;

    bool return_closest_points(int, NEWMESH::Pt&, NEWMESH::Pt&,NEWMESH::Pt&,const NEWMESH::Pt &, int &, int &, int &) const ;


    //-----------FOR ACCESS USING ITERATORS ----------------//

    std::vector<std::shared_ptr<Mpoint> >::const_iterator vbegin() const { return _points.begin();};   /// const to prevent reset of pointers - note this means _points can be modified
    std::vector<std::shared_ptr<Mpoint> >::const_iterator vend() const {return _points.end();};

    std::vector<NEWMESH::Triangle>::iterator tbegin(){ return _triangles.begin();};
    std::vector<NEWMESH::Triangle>::const_iterator tbegin() const{ return _triangles.begin();};
    std::vector<NEWMESH::Triangle>::iterator tend(){ return _triangles.end();};
    std::vector<NEWMESH::Triangle>::const_iterator tend() const { return _triangles.end();};

    std::vector<int>::const_iterator nbegin(const int &i) const{ return _points[i]->_nID.begin();}; // neighbours of a particular vertex, access only
    std::vector<int>::const_iterator nend(const int &i) const { return _points[i]->_nID.end();};

    std::vector<int>::const_iterator tIDbegin(const int &i) const{ return _points[i]->_trID.begin();}; //faces associated with a particular vertex
    std::vector<int>::const_iterator tIDend(const int &i) const { return _points[i]->_trID.end();};

    /////-----------ASSIGNMENT --------------------------------//
    inline void set_coord(const int& i,const Pt& p){_points[i]->set_coord(p);}
    void set_pvalue(const unsigned int& i,const float& val, const int dim=0);
    void set_tvalue(const unsigned int& i,const float& val, const int dim=0);
    void set_pvalues(const std::vector<int>& ids,const float& val,const int dim=0);
    void set_pvalues(const NEWMAT::Matrix& M,const bool appendFieldData=false);


    void estimate_normals();

    void initialize_pvalues(const int dim=1,const bool appendFieldData=false); // initialize new data array
    void initialize_tvalues(const int dim=1,const bool appendFieldData=false); // initialize new tvalue data array

    int step_sign(const int& vertind,const Pt& step)const;

    Pt estimate_origin();
    ////////////////  LOAD FUNCTIONS /////////////////
    ////////////////  appendFieldData refers only to pvalue/tvalue fields. As such if you are reading a .func or a .shape following a .surf appendFieldData=false. ////////

    void load(const std::string& filename, const bool loadSurfaceData=true, const bool appendFieldData=false); /////////// ASSUME EITHER SURFACE FILE OR DATA FILE NOT BOTH //////////////
    void load_gifti(const std::string& filename, const bool loadSurfaceData=true, const bool appendFieldData=false);
    void load_ascii(const std::string& filename, const bool loadSurfaceData=true, const bool appendFieldData=false);  /// loads either surface or data
    void load_ascii_file(const std::string& filename); /// loads surface & data
    void load_vtk(const std::string& filename); // cannot provide field data - surface only
    void load_matrix(const std::string& filename,const int& type) ;// cannot provide field data - surface only (also reads .dpv files)


    ///////////////////// WRITE FUNCTIONS ///////////////////////////////////////////
    void save(const std::string& filename)const;
    void save_ascii(const std::string& s)const;
    void save_gifti(const std::string& s) const;
    void save_vtk(const std::string& s)const;
    void save_dpv(const std::string& s)const;
    void save_matrix(const std::string& s)const;

    // void save_gifti(const string& s,const int& type=GIFTI_ENCODING_B64GZ) const;


    void print(const std::string& filename);

    //  ostream& operator <<(ostream& flot);
  };

  /// generic function for projecting points
  bool SameSide(const NEWMESH::Pt&, const NEWMESH::Pt&, const NEWMESH::Pt&, const NEWMESH::Pt&);
  bool PointInTriangle(const NEWMESH::Pt& , const NEWMESH::Pt&, const NEWMESH::Pt&, const NEWMESH::Pt&);

  NEWMESH::Pt projectPoint(const NEWMESH::Pt &, const NEWMESH::Pt &, const NEWMESH::Pt &, const NEWMESH::Pt &, NEWMESH::Pt &);
  void projectPoint(const NEWMESH::Pt &, const NEWMESH::Pt &, const NEWMESH::Pt &, NEWMESH::Pt &);
  void projectVector(const NEWMESH::Pt &, const NEWMESH::Pt &, const NEWMESH::Pt &, const NEWMESH::Pt &, NEWMESH::Pt &);


}

#endif
